/*!
 * @antv/g-plugin-canvaskit-renderer
 * @description A G plugin of renderer implementation with CanvasKit
 * @version 2.1.23
 * @date 5/30/2025, 2:55:37 AM
 * @author AntVis
 * @docs https://g.antv.antgroup.com/
 */
'use strict';

var _defineProperty = require('@babel/runtime/helpers/defineProperty');
var _objectSpread = require('@babel/runtime/helpers/objectSpread2');
var _classCallCheck = require('@babel/runtime/helpers/classCallCheck');
var _createClass = require('@babel/runtime/helpers/createClass');
var _callSuper = require('@babel/runtime/helpers/callSuper');
var _inherits = require('@babel/runtime/helpers/inherits');
var gLite = require('@antv/g-lite');
var _slicedToArray = require('@babel/runtime/helpers/slicedToArray');
var _regeneratorRuntime = require('@babel/runtime/helpers/regeneratorRuntime');
var _asyncToGenerator = require('@babel/runtime/helpers/asyncToGenerator');
var util = require('@antv/util');
var glMatrix = require('gl-matrix');
var _toConsumableArray = require('@babel/runtime/helpers/toConsumableArray');

var CircleRenderer = /*#__PURE__*/function () {
  function CircleRenderer() {
    _classCallCheck(this, CircleRenderer);
  }
  return _createClass(CircleRenderer, [{
    key: "render",
    value: function render(object, context) {
      var canvas = context.canvas,
        strokePaint = context.strokePaint,
        fillPaint = context.fillPaint,
        shadowFillPaint = context.shadowFillPaint,
        shadowStrokePaint = context.shadowStrokePaint;
      var _ref = object.parsedStyle,
        cx = _ref.cx,
        cy = _ref.cy,
        r = _ref.r,
        shadowOffsetX = _ref.shadowOffsetX,
        shadowOffsetY = _ref.shadowOffsetY;
      if (shadowFillPaint || shadowStrokePaint) {
        canvas.drawCircle(cx + (shadowOffsetX || 0) / 2, cy + (shadowOffsetY || 0) / 2, r, shadowFillPaint || shadowStrokePaint);
      }
      if (fillPaint) {
        canvas.drawCircle(cx, cy, r, fillPaint);
      }
      if (strokePaint) {
        canvas.drawCircle(cx, cy, r, strokePaint);
      }
    }
  }]);
}();

/**
 * @see https://fiddle.skia.org/c/@Canvas_drawOval
 * @example
 * SkRect bounds = SkRect::MakeWH(80, 70);
 */
var EllipseRenderer = /*#__PURE__*/function () {
  function EllipseRenderer() {
    _classCallCheck(this, EllipseRenderer);
  }
  return _createClass(EllipseRenderer, [{
    key: "render",
    value: function render(object, context) {
      var canvas = context.canvas,
        strokePaint = context.strokePaint,
        fillPaint = context.fillPaint,
        shadowFillPaint = context.shadowFillPaint,
        shadowStrokePaint = context.shadowStrokePaint;
      var _ref = object.parsedStyle,
        cx = _ref.cx,
        cy = _ref.cy,
        rx = _ref.rx,
        ry = _ref.ry,
        shadowOffsetX = _ref.shadowOffsetX,
        shadowOffsetY = _ref.shadowOffsetY;
      if (shadowFillPaint || shadowStrokePaint) {
        canvas.drawOval([cx - rx + (shadowOffsetX || 0), cy - ry + (shadowOffsetY || 0), cx + rx + (shadowOffsetX || 0), cy + ry + (shadowOffsetY || 0)], shadowFillPaint || shadowStrokePaint);
      }
      if (fillPaint) {
        canvas.drawOval([cx - rx, cy - ry, cx + rx, cy + ry], fillPaint);
      }
      if (strokePaint) {
        canvas.drawOval([cx - rx, cy - ry, cx + rx, cy + ry], strokePaint);
      }
    }
  }]);
}();

/**
 * @see https://docs.flutter.dev/development/platform-integration/web-images#flutter-renderers-on-the-web
 * @see https://stackoverflow.com/questions/71418851/load-images-for-canvaskit-wasm-on-node-js-environment
 * @see https://fiddle.skia.org/c/@Canvas_drawImage
 * @see https://github.com/google/skia/blob/4ff73144c35b993907a6e3738a7be81c0681e504/modules/canvaskit/tests/bazel/core_test.js#L104
 */
var ImageRenderer = /*#__PURE__*/function () {
  function ImageRenderer(context) {
    _classCallCheck(this, ImageRenderer);
    this.context = context;
  }
  return _createClass(ImageRenderer, [{
    key: "render",
    value: function render(object, context) {
      var _getContext = this.context.contextService.getContext(),
        surface = _getContext.surface,
        CanvasKit = _getContext.CanvasKit;
      var canvas = context.canvas;
      var _ref = object.parsedStyle,
        x = _ref.x,
        y = _ref.y,
        width = _ref.width,
        height = _ref.height,
        src = _ref.src,
        fillOpacity = _ref.fillOpacity,
        opacity = _ref.opacity;
      var imageCache = this.context.imagePool.getImageSync(src, object);
      var image = imageCache === null || imageCache === void 0 ? void 0 : imageCache.img;
      var iw = width;
      var ih = height;
      if (image) {
        iw || (iw = image.width);
        ih || (ih = image.height);
        var decoded = surface.makeImageFromTextureSource(image
        // {
        //   alphaType: CanvasKit.AlphaType.Unpremul,
        //   colorType: CanvasKit.ColorType.RGBA_8888,
        //   colorSpace: CanvasKit.ColorSpace.SRGB,
        //   width: iw,
        //   height: ih,
        // }
        );
        var srcRect = CanvasKit.XYWHRect(0, 0, image.width, image.height);
        var destRect = CanvasKit.XYWHRect(x, y, iw, ih);
        var fillPaint = new CanvasKit.Paint();
        fillPaint.setAntiAlias(true);
        fillPaint.setStyle(CanvasKit.PaintStyle.Fill);
        fillPaint.setAlphaf(fillOpacity * opacity);

        // @see https://github.com/google/skia/blob/4ff73144c35b993907a6e3738a7be81c0681e504/modules/canvaskit/tests/core.spec.js#L864
        canvas.drawImageRectOptions(decoded, srcRect, destRect, CanvasKit.FilterMode.Linear, CanvasKit.MipmapMode.None, fillPaint);
        fillPaint["delete"]();
      }
    }
  }]);
}();

/**
 * @see https://fiddle.skia.org/c/@Canvas_drawLine
 */
var LineRenderer = /*#__PURE__*/function () {
  function LineRenderer() {
    _classCallCheck(this, LineRenderer);
  }
  return _createClass(LineRenderer, [{
    key: "render",
    value: function render(object, context) {
      var canvas = context.canvas,
        strokePaint = context.strokePaint,
        shadowStrokePaint = context.shadowStrokePaint;
      var _ref = object.parsedStyle,
        shadowOffsetX = _ref.shadowOffsetX,
        shadowOffsetY = _ref.shadowOffsetY,
        x1 = _ref.x1,
        y1 = _ref.y1,
        x2 = _ref.x2,
        y2 = _ref.y2,
        markerStart = _ref.markerStart,
        markerEnd = _ref.markerEnd,
        markerStartOffset = _ref.markerStartOffset,
        markerEndOffset = _ref.markerEndOffset;
      var startOffsetX = 0;
      var startOffsetY = 0;
      var endOffsetX = 0;
      var endOffsetY = 0;
      var rad = 0;
      var x;
      var y;
      if (markerStart && gLite.isDisplayObject(markerStart) && markerStartOffset) {
        x = x2 - x1;
        y = y2 - y1;
        rad = Math.atan2(y, x);
        startOffsetX = Math.cos(rad) * (markerStartOffset || 0);
        startOffsetY = Math.sin(rad) * (markerStartOffset || 0);
      }
      if (markerEnd && gLite.isDisplayObject(markerEnd) && markerEndOffset) {
        x = x1 - x2;
        y = y1 - y2;
        rad = Math.atan2(y, x);
        endOffsetX = Math.cos(rad) * (markerEndOffset || 0);
        endOffsetY = Math.sin(rad) * (markerEndOffset || 0);
      }
      if (shadowStrokePaint) {
        canvas.drawLine(x1 + (shadowOffsetX || 0) / 2 + startOffsetX, y1 + (shadowOffsetY || 0) / 2 + startOffsetY, x2 + (shadowOffsetX || 0) / 2 + endOffsetX, y2 + (shadowOffsetY || 0) / 2 + endOffsetY, shadowStrokePaint);
      }
      if (strokePaint) {
        canvas.drawLine(x1 + startOffsetX, y1 + startOffsetY, x2 + endOffsetX, y2 + endOffsetY, strokePaint);
      }
    }
  }]);
}();

/**
 * Generate SkPath from G's Path.
 * @see https://api.skia.org/classSkPath.html
 */
function generateSkPath(CanvasKit, object) {
  var skPath = new CanvasKit.Path();
  var _object$parsedStyle = object.parsedStyle,
    d = _object$parsedStyle.d,
    markerStart = _object$parsedStyle.markerStart,
    markerEnd = _object$parsedStyle.markerEnd,
    markerStartOffset = _object$parsedStyle.markerStartOffset,
    markerEndOffset = _object$parsedStyle.markerEndOffset;
  var startOffsetX = 0;
  var startOffsetY = 0;
  var endOffsetX = 0;
  var endOffsetY = 0;
  var rad = 0;
  var x;
  var y;
  if (markerStart && gLite.isDisplayObject(markerStart) && markerStartOffset) {
    var _getStartTangent = markerStart.parentNode.getStartTangent(),
      _getStartTangent2 = _slicedToArray(_getStartTangent, 2),
      p1 = _getStartTangent2[0],
      p2 = _getStartTangent2[1];
    x = p1[0] - p2[0];
    y = p1[1] - p2[1];
    rad = Math.atan2(y, x);
    startOffsetX = Math.cos(rad) * (markerStartOffset || 0);
    startOffsetY = Math.sin(rad) * (markerStartOffset || 0);
  }
  if (markerEnd && gLite.isDisplayObject(markerEnd) && markerEndOffset) {
    var _getEndTangent = markerEnd.parentNode.getEndTangent(),
      _getEndTangent2 = _slicedToArray(_getEndTangent, 2),
      _p = _getEndTangent2[0],
      _p2 = _getEndTangent2[1];
    x = _p[0] - _p2[0];
    y = _p[1] - _p2[1];
    rad = Math.atan2(y, x);
    endOffsetX = Math.cos(rad) * (markerEndOffset || 0);
    endOffsetY = Math.sin(rad) * (markerEndOffset || 0);
  }
  var absolutePath = d.absolutePath,
    segments = d.segments;
  for (var i = 0; i < absolutePath.length; i++) {
    var params = absolutePath[i]; // eg. M 100 200
    var command = params[0];
    var nextSegment = absolutePath[i + 1];
    var useStartOffset = i === 0 && (startOffsetX !== 0 || startOffsetY !== 0);
    var useEndOffset = (i === absolutePath.length - 1 || nextSegment && (nextSegment[0] === 'M' || nextSegment[0] === 'Z')) && endOffsetX !== 0 && endOffsetY !== 0;
    switch (command) {
      case 'M':
        // Use start marker offset
        if (useStartOffset) {
          skPath.moveTo(params[1] + startOffsetX, params[2] + startOffsetY);
          skPath.lineTo(params[1], params[2]);
        } else {
          skPath.moveTo(params[1], params[2]);
        }
        break;
      case 'L':
        if (useEndOffset) {
          skPath.lineTo(params[1] + endOffsetX, params[2] + endOffsetY);
        } else {
          skPath.lineTo(params[1], params[2]);
        }
        break;
      case 'Q':
        skPath.quadTo(params[1], params[2], params[3], params[4]);
        if (useEndOffset) {
          skPath.lineTo(params[3] + endOffsetX, params[4] + endOffsetY);
        }
        break;
      case 'C':
        skPath.cubicTo(params[1], params[2], params[3], params[4], params[5], params[6]);
        if (useEndOffset) {
          skPath.lineTo(params[5] + endOffsetX, params[6] + endOffsetY);
        }
        break;
      case 'A':
        {
          var arcParams = segments[i].arcParams;
          var rx = arcParams.rx,
            ry = arcParams.ry,
            sweepFlag = arcParams.sweepFlag;
          var largeArcFlag = params[4];
          skPath.arcToRotated(rx, ry, params[3], !largeArcFlag,
          // useSmallArc
          !!(1 - sweepFlag), params[6], params[7]);
          if (useEndOffset) {
            skPath.lineTo(params[6] + endOffsetX, params[7] + endOffsetY);
          }
          break;
        }
      case 'Z':
        skPath.close();
        break;
    }
  }
  return skPath;
}

/**
 * @see https://fiddle.skia.org/c/@Canvas_drawPath
 */
var PathRenderer = /*#__PURE__*/function () {
  function PathRenderer(context) {
    _classCallCheck(this, PathRenderer);
    this.context = context;
  }
  return _createClass(PathRenderer, [{
    key: "render",
    value: function render(object, context) {
      var _getContext = this.context.contextService.getContext(),
        CanvasKit = _getContext.CanvasKit;
      var canvas = context.canvas,
        fillPaint = context.fillPaint,
        strokePaint = context.strokePaint,
        shadowFillPaint = context.shadowFillPaint,
        shadowStrokePaint = context.shadowStrokePaint;
      var _object$parsedStyle2 = object.parsedStyle,
        shadowOffsetX = _object$parsedStyle2.shadowOffsetX,
        shadowOffsetY = _object$parsedStyle2.shadowOffsetY;
      var skPath = generateSkPath(CanvasKit, object);
      if (shadowFillPaint || shadowStrokePaint) {
        var shadowPath = skPath.copy();
        shadowPath.transform(glMatrix.mat3.fromTranslation(glMatrix.mat3.create(), [(shadowOffsetX || 0) / 2, (shadowOffsetY || 0) / 2]));
        canvas.drawPath(shadowPath, fillPaint ? shadowFillPaint : shadowStrokePaint);
      }
      if (fillPaint) {
        canvas.drawPath(skPath, fillPaint);
      }
      if (strokePaint) {
        canvas.drawPath(skPath, strokePaint);
      }
    }
  }]);
}();

/**
 * @see https://fiddle.skia.org/c/@Path_addPoly
 */
var PolygonRenderer = /*#__PURE__*/function () {
  function PolygonRenderer(context) {
    _classCallCheck(this, PolygonRenderer);
    this.context = context;
  }
  return _createClass(PolygonRenderer, [{
    key: "render",
    value: function render(object, context) {
      var _getContext = this.context.contextService.getContext(),
        CanvasKit = _getContext.CanvasKit;
      var canvas = context.canvas,
        fillPaint = context.fillPaint,
        strokePaint = context.strokePaint,
        shadowFillPaint = context.shadowFillPaint,
        shadowStrokePaint = context.shadowStrokePaint;
      var _ref = object.parsedStyle,
        shadowOffsetX = _ref.shadowOffsetX,
        shadowOffsetY = _ref.shadowOffsetY,
        points = _ref.points.points,
        markerStart = _ref.markerStart,
        markerEnd = _ref.markerEnd,
        markerStartOffset = _ref.markerStartOffset,
        markerEndOffset = _ref.markerEndOffset;
      var length = points.length;
      var startOffsetX = 0;
      var startOffsetY = 0;
      var endOffsetX = 0;
      var endOffsetY = 0;
      var rad = 0;
      var x;
      var y;
      if (markerStart && gLite.isDisplayObject(markerStart) && markerStartOffset) {
        x = points[1][0] - points[0][0];
        y = points[1][1] - points[0][1];
        rad = Math.atan2(y, x);
        startOffsetX = Math.cos(rad) * (markerStartOffset || 0);
        startOffsetY = Math.sin(rad) * (markerStartOffset || 0);
      }
      if (markerEnd && gLite.isDisplayObject(markerEnd) && markerEndOffset) {
        x = points[length - 2][0] - points[length - 1][0];
        y = points[length - 2][1] - points[length - 1][1];
        rad = Math.atan2(y, x);
        endOffsetX = Math.cos(rad) * (markerEndOffset || 0);
        endOffsetY = Math.sin(rad) * (markerEndOffset || 0);
      }
      var formattedPoints = points.map(function (_ref2, i) {
        var _ref3 = _slicedToArray(_ref2, 2),
          x = _ref3[0],
          y = _ref3[1];
        var offsetX = 0;
        var offsetY = 0;
        if (i === 0) {
          offsetX = startOffsetX;
          offsetY = startOffsetY;
        } else if (i === length - 1) {
          offsetX = endOffsetX;
          offsetY = endOffsetY;
        }
        return [x + offsetX, y + offsetY];
      }).reduce(function (prev, cur) {
        return prev.concat(cur);
      }, []);
      if (shadowFillPaint || shadowStrokePaint) {
        var _path = new CanvasKit.Path();
        _path.addPoly(formattedPoints.map(function (x, i) {
          return x + (i % 2 === 0 ? (shadowOffsetX || 0) / 2 : (shadowOffsetY || 0) / 2);
        }), true);
        canvas.drawPath(_path, fillPaint ? shadowFillPaint : shadowStrokePaint);
      }
      var path = new CanvasKit.Path();
      path.addPoly(formattedPoints, true);
      if (fillPaint) {
        canvas.drawPath(path, fillPaint);
      }
      if (strokePaint) {
        canvas.drawPath(path, strokePaint);
      }
    }
  }]);
}();

/**
 * @see https://fiddle.skia.org/c/@Path_addPoly
 */
var PolylineRenderer = /*#__PURE__*/function () {
  function PolylineRenderer(context) {
    _classCallCheck(this, PolylineRenderer);
    this.context = context;
  }
  return _createClass(PolylineRenderer, [{
    key: "render",
    value: function render(object, context) {
      var _getContext = this.context.contextService.getContext(),
        CanvasKit = _getContext.CanvasKit;
      var canvas = context.canvas,
        strokePaint = context.strokePaint,
        shadowStrokePaint = context.shadowStrokePaint;
      var _ref = object.parsedStyle,
        shadowOffsetX = _ref.shadowOffsetX,
        shadowOffsetY = _ref.shadowOffsetY,
        points = _ref.points.points,
        markerStart = _ref.markerStart,
        markerEnd = _ref.markerEnd,
        markerStartOffset = _ref.markerStartOffset,
        markerEndOffset = _ref.markerEndOffset;
      var length = points.length;
      var startOffsetX = 0;
      var startOffsetY = 0;
      var endOffsetX = 0;
      var endOffsetY = 0;
      var rad = 0;
      var x;
      var y;
      if (markerStart && gLite.isDisplayObject(markerStart) && markerStartOffset) {
        x = points[1][0] - points[0][0];
        y = points[1][1] - points[0][1];
        rad = Math.atan2(y, x);
        startOffsetX = Math.cos(rad) * (markerStartOffset || 0);
        startOffsetY = Math.sin(rad) * (markerStartOffset || 0);
      }
      if (markerEnd && gLite.isDisplayObject(markerEnd) && markerEndOffset) {
        x = points[length - 2][0] - points[length - 1][0];
        y = points[length - 2][1] - points[length - 1][1];
        rad = Math.atan2(y, x);
        endOffsetX = Math.cos(rad) * (markerEndOffset || 0);
        endOffsetY = Math.sin(rad) * (markerEndOffset || 0);
      }
      var formattedPoints = points.map(function (_ref2, i) {
        var _ref3 = _slicedToArray(_ref2, 2),
          x = _ref3[0],
          y = _ref3[1];
        var offsetX = 0;
        var offsetY = 0;
        if (i === 0) {
          offsetX = startOffsetX;
          offsetY = startOffsetY;
        } else if (i === length - 1) {
          offsetX = endOffsetX;
          offsetY = endOffsetY;
        }
        return [x + offsetX, y + offsetY];
      }).reduce(function (prev, cur) {
        return prev.concat(cur);
      }, []);
      if (shadowStrokePaint) {
        var _path = new CanvasKit.Path();
        _path.addPoly(formattedPoints.map(function (x, i) {
          return x + (i % 2 === 0 ? (shadowOffsetX || 0) / 2 : (shadowOffsetY || 0) / 2);
        }), false);
        canvas.drawPath(_path, shadowStrokePaint);
      }
      var path = new CanvasKit.Path();
      path.addPoly(formattedPoints, false);
      if (strokePaint) {
        canvas.drawPath(path, strokePaint);
      }
    }
  }]);
}();

/**
 * should account for round rect(RRect)
 * @see https://fiddle.skia.org/c/@Canvas_drawRoundRect
 *
 * An RRect (rectangle with rounded corners) is represented by 12 floats. In order, the floats
 * correspond to left, top, right, bottom and then in pairs, the radiusX, radiusY for upper-left,
 * upper-right, lower-right, lower-left. See RRect.h for more.
 */
var RectRenderer = /*#__PURE__*/function () {
  function RectRenderer() {
    _classCallCheck(this, RectRenderer);
  }
  return _createClass(RectRenderer, [{
    key: "render",
    value: function render(object, context) {
      var canvas = context.canvas,
        strokePaint = context.strokePaint,
        fillPaint = context.fillPaint,
        shadowFillPaint = context.shadowFillPaint,
        shadowStrokePaint = context.shadowStrokePaint;
      var _ref = object.parsedStyle,
        x = _ref.x,
        y = _ref.y,
        width = _ref.width,
        height = _ref.height,
        radius = _ref.radius,
        shadowOffsetX = _ref.shadowOffsetX,
        shadowOffsetY = _ref.shadowOffsetY;
      var _radius$map = radius.map(function (r) {
          return util.clamp(r, 0, Math.min(Math.abs(width) / 2, Math.abs(height) / 2));
        }),
        _radius$map2 = _slicedToArray(_radius$map, 4),
        r1 = _radius$map2[0],
        r2 = _radius$map2[1],
        r3 = _radius$map2[2],
        r4 = _radius$map2[3];
      var flipY = width < 0;
      var flipX = height < 0;
      var tlr;
      var trr;
      var brr;
      var blr;
      if (!flipX && !flipY) {
        tlr = r1;
        trr = r2;
        brr = r3;
        blr = r4;
      } else if (flipX && flipY) {
        tlr = r3;
        trr = r4;
        brr = r1;
        blr = r2;
      } else if (flipX && !flipY) {
        tlr = r4;
        trr = r3;
        brr = r2;
        blr = r1;
      } else if (!flipX && flipY) {
        tlr = r2;
        trr = r1;
        brr = r4;
        blr = r3;
      }
      var rrect = [x, y, x + width, y + height, tlr, tlr, trr, trr, brr, brr, blr, blr];
      if (shadowFillPaint || shadowStrokePaint) {
        canvas.drawRRect([x + (shadowOffsetX || 0) / 2, y + (shadowOffsetY || 0) / 2, x + width + (shadowOffsetX || 0) / 2, y + height + (shadowOffsetY || 0) / 2, tlr, tlr, trr, trr, brr, brr, blr, blr], fillPaint ? shadowFillPaint : shadowStrokePaint);
      }
      if (fillPaint) {
        canvas.drawRRect(rrect, fillPaint);
      }
      if (strokePaint) {
        canvas.drawRRect(rrect, strokePaint);
      }
    }
  }]);
}();

function color2CanvaskitColor(CanvasKit, color) {
  var parsedColor = gLite.parseColor(color);
  return CanvasKit.Color4f(Number(parsedColor.r) / 255, Number(parsedColor.g) / 255, Number(parsedColor.b) / 255, Number(parsedColor.alpha));
}

/**
 * One of the biggest features that CanvasKit offers over the HTML Canvas API is paragraph shaping.
 * To use text your applicatoin, supply a font file and use Promise.all to run your code when both CanvasKit and the font file are ready.
 *
 * @see https://skia.org/docs/user/modules/quickstart/#text-shaping
 * @see https://fiddle.skia.org/c/@Canvas_drawText
 *
 * Depend on volatile Google Fonts:
 * @see https://github.com/flutter/flutter/issues/85793
 *
 * CJK Fonts:
 * @see https://github.com/flutter/flutter/issues/77212
 * @see https://github.com/flutter/flutter/issues/53897
 *
 * Emoji:
 * @see https://github.com/flutter/flutter/issues/76248
 */
var TextRenderer = /*#__PURE__*/function () {
  function TextRenderer(context, fontLoader) {
    _classCallCheck(this, TextRenderer);
    this.context = context;
    this.fontLoader = fontLoader;
  }
  return _createClass(TextRenderer, [{
    key: "render",
    value: function render(object, context) {
      var _this = this;
      var _getContext = this.context.contextService.getContext(),
        CanvasKit = _getContext.CanvasKit;
      var canvas = context.canvas;
      var _ref = object.parsedStyle,
        text = _ref.text,
        fontSize = _ref.fontSize,
        _ref$fontFamily = _ref.fontFamily,
        fontFamilies = _ref$fontFamily === void 0 ? '' : _ref$fontFamily,
        fontWeight = _ref.fontWeight,
        textAlign = _ref.textAlign,
        textBaseline = _ref.textBaseline,
        letterSpacing = _ref.letterSpacing,
        fill = _ref.fill,
        wordWrap = _ref.wordWrap,
        wordWrapWidth = _ref.wordWrapWidth,
        x = _ref.x,
        y = _ref.y,
        dx = _ref.dx,
        dy = _ref.dy,
        alongPath = _ref.alongPath,
        maxLines = _ref.maxLines,
        ellipsis = _ref.ellipsis,
        decorationLine = _ref.decorationLine,
        decorationThickness = _ref.decorationThickness,
        decorationStyle = _ref.decorationStyle,
        decorationColor = _ref.decorationColor,
        direction = _ref.direction,
        backgroundColor = _ref.backgroundColor,
        foregroundColor = _ref.foregroundColor,
        wordSpacing = _ref.wordSpacing,
        disableHinting = _ref.disableHinting,
        shadows = _ref.shadows,
        halfLeading = _ref.halfLeading,
        fontFeatures = _ref.fontFeatures,
        strutStyle = _ref.strutStyle,
        heightMultiplier = _ref.heightMultiplier;
      var TEXT_ALIGN_MAP = {
        left: CanvasKit.TextAlign.Left,
        center: CanvasKit.TextAlign.Center,
        middle: CanvasKit.TextAlign.Center,
        right: CanvasKit.TextAlign.Right,
        end: CanvasKit.TextAlign.End,
        start: CanvasKit.TextAlign.Start
      };
      var TEXT_BASELINE_MAP = {
        alphabetic: CanvasKit.TextBaseline.Alphabetic,
        bottom: undefined,
        hanging: undefined,
        ideographic: CanvasKit.TextBaseline.Ideographic,
        middle: undefined,
        top: undefined
      };

      // fontFamily
      var loadedFontFamilies = [];
      var loadedFontData = [];
      var loadedTypefaces = [];
      fontFamilies.split(',').forEach(function (fontFamily) {
        loadedFontFamilies.push(fontFamily.trim());
        loadedFontData.push(_this.fontLoader.getFontBuffer(fontFamily.trim()));
        loadedTypefaces.push(_this.fontLoader.getTypeface(fontFamily.trim()));
      });
      if (alongPath) {
        var skPath = new CanvasKit.Path();
        var _ref2 = alongPath.parsedStyle,
          parsedPath = _ref2.d;
        var absolutePath = parsedPath.absolutePath,
          segments = parsedPath.segments;
        for (var i = 0; i < absolutePath.length; i++) {
          var params = absolutePath[i]; // eg. M 100 200
          var command = params[0];
          // V,H,S,T 都在前面被转换成标准形式
          switch (command) {
            case 'M':
              skPath.moveTo(params[1], params[2]);
              break;
            case 'L':
              skPath.lineTo(params[1], params[2]);
              break;
            case 'Q':
              skPath.quadTo(params[1], params[2], params[3], params[4]);
              break;
            case 'C':
              skPath.cubicTo(params[1], params[2], params[3], params[4], params[5], params[6]);
              break;
            case 'A':
              {
                var arcParams = segments[i].arcParams;
                var rx = arcParams.rx,
                  ry = arcParams.ry,
                  sweepFlag = arcParams.sweepFlag;
                var largeArcFlag = params[4];
                skPath.arcToRotated(rx, ry, params[3], !largeArcFlag,
                // useSmallArc
                !!(1 - sweepFlag), params[6], params[7]);
                break;
              }
            case 'Z':
              skPath.close();
              break;
          }
        }
        var textPaint = new CanvasKit.Paint();
        textPaint.setAntiAlias(true);
        textPaint.setStyle(CanvasKit.PaintStyle.Fill);
        if (gLite.isCSSRGB(fill) && !fill.isNone) {
          textPaint.setColor(CanvasKit.Color4f(Number(fill.r) / 255, Number(fill.g) / 255, Number(fill.b) / 255, Number(fill.alpha)));
        }
        var skFont = new CanvasKit.Font(loadedTypefaces[0], fontSize);
        var textblob = CanvasKit.TextBlob.MakeOnPath(text, skPath, skFont);
        canvas.drawTextBlob(textblob, 0, 0, textPaint);
      } else {
        var DECORATION_MAP = {
          none: CanvasKit.NoDecoration,
          underline: CanvasKit.UnderlineDecoration,
          overline: CanvasKit.OverlineDecoration,
          'line-through': CanvasKit.LineThroughDecoration
        };
        var DECORATION_STYLE_MAP = {
          solid: CanvasKit.DecorationStyle.Solid,
          "double": CanvasKit.DecorationStyle.Double,
          dotted: CanvasKit.DecorationStyle.Dotted,
          dashed: CanvasKit.DecorationStyle.Dashed,
          wavy: CanvasKit.DecorationStyle.Wavy
        };
        var DIRECTION_MAP = {
          ltr: CanvasKit.TextDirection.LTR,
          rtl: CanvasKit.TextDirection.RTL
        };
        var paraStyle = new CanvasKit.ParagraphStyle({
          textStyle: {
            backgroundColor: color2CanvaskitColor(CanvasKit, backgroundColor),
            color: CanvasKit.Color4f(Number(fill.r) / 255, Number(fill.g) / 255, Number(fill.b) / 255, Number(fill.alpha)),
            decoration: DECORATION_MAP[decorationLine || 'none'],
            decorationColor: color2CanvaskitColor(CanvasKit, decorationColor),
            decorationThickness: decorationThickness,
            decorationStyle: DECORATION_STYLE_MAP[decorationStyle || 'solid'],
            fontFamilies: loadedFontFamilies,
            fontFeatures: fontFeatures,
            fontSize: fontSize,
            fontStyle: {
              weight: {
                value: Number(fontWeight)
              }
              // width?: FontWidth;
              // slant?: FontSlant;
            },
            foregroundColor: color2CanvaskitColor(CanvasKit, foregroundColor || object.getAttribute('fill')),
            heightMultiplier: heightMultiplier,
            halfLeading: halfLeading,
            letterSpacing: letterSpacing,
            // locale?: string;
            shadows: (shadows || []).map(function (_ref3) {
              var color = _ref3.color,
                offset = _ref3.offset,
                blurRadius = _ref3.blurRadius;
              return {
                color: color2CanvaskitColor(CanvasKit, color),
                offset: offset,
                blurRadius: blurRadius
              };
            }),
            textBaseline: TEXT_BASELINE_MAP[textBaseline],
            wordSpacing: wordSpacing
          },
          textAlign: TEXT_ALIGN_MAP[textAlign],
          disableHinting: disableHinting,
          ellipsis: ellipsis,
          // heightMultiplier,
          maxLines: maxLines,
          strutStyle: strutStyle,
          textDirection: DIRECTION_MAP[direction || 'ltr']
          // textHeightBehavior?: TextHeightBehavior;
        });

        // use cached font manager
        var fontMgr = this.fontLoader.getOrCreateFontMgr(CanvasKit, loadedFontFamilies);
        var builder = CanvasKit.ParagraphBuilder.Make(paraStyle, fontMgr);
        builder.addText(text);
        var paragraph = builder.build();
        if (wordWrap) {
          // width in pixels to use when wrapping text
          paragraph.layout(wordWrapWidth);
        } else {
          paragraph.layout(text.length * fontSize);
        }

        // account for textBaseline
        var paragraphHeight = paragraph.getHeight();
        var paragraphMaxWidth = paragraph.getMaxWidth();
        var offsetX = 0;
        // handle horizontal text align
        if (textAlign === 'center' || textAlign === 'middle') {
          offsetX -= paragraphMaxWidth / 2;
        } else if (textAlign === 'right' || textAlign === 'end') {
          offsetX -= paragraphMaxWidth;
        }
        var linePositionY = 0;
        // handle vertical text baseline
        if (textBaseline === 'middle') {
          linePositionY = -paragraphHeight / 2;
        } else if (textBaseline === 'bottom' || textBaseline === 'alphabetic' || textBaseline === 'ideographic') {
          linePositionY = -paragraphHeight;
        } else if (textBaseline === 'top' || textBaseline === 'hanging') {
          linePositionY = 0;
        }
        canvas.drawParagraph(paragraph, offsetX + x + dx, linePositionY + y + dy);
        paragraph["delete"]();
        builder["delete"]();
      }
    }
  }]);
}();

/**
 * @see https://skia.org/docs/user/modules/quickstart/
 */
var CanvaskitRendererPlugin = /*#__PURE__*/function () {
  function CanvaskitRendererPlugin(canvaskitRendererPluginOptions, rendererContributionFactory, fontLoader) {
    _classCallCheck(this, CanvaskitRendererPlugin);
    this.destroyed = false;
    this.animations = [];
    this.particlesList = [];
    /**
     * This stack is only used by clipPath for now.
     */
    this.restoreStack = [];
    this.canvaskitRendererPluginOptions = canvaskitRendererPluginOptions;
    this.rendererContributionFactory = rendererContributionFactory;
    this.fontLoader = fontLoader;
  }
  return _createClass(CanvaskitRendererPlugin, [{
    key: "playAnimation",
    value: function playAnimation(name, jsonStr, bounds, assets) {
      var canvasKitContext = this.context.contextService.getContext();
      var CanvasKit = canvasKitContext.CanvasKit;
      var animation = CanvasKit.MakeManagedAnimation(jsonStr, assets);
      var duration = animation.duration() * 1000;
      var size = animation.size();
      this.animations.push({
        name: name,
        jsonStr: jsonStr,
        bounds: bounds || CanvasKit.LTRBRect(0, 0, size[0], size[1]),
        assets: assets,
        duration: duration,
        size: size,
        animation: animation
      });
      return animation;
    }
  }, {
    key: "createParticles",
    value: function createParticles(jsonStr, onFrame, assets) {
      var canvasKitContext = this.context.contextService.getContext();
      var CanvasKit = canvasKitContext.CanvasKit;
      var particles = CanvasKit.MakeParticles(jsonStr, assets);
      this.particlesList.push({
        particles: particles,
        onFrame: onFrame
      });
      return particles;
    }
  }, {
    key: "apply",
    value: function apply(context) {
      var _this = this;
      this.context = context;
      var renderingService = context.renderingService,
        renderingContext = context.renderingContext;
      renderingService.hooks.init.tap(CanvaskitRendererPlugin.tag, function () {
        var canvasKitContext = _this.context.contextService.getContext();
        var surface = canvasKitContext.surface,
          CanvasKit = canvasKitContext.CanvasKit;
        var fonts = _this.canvaskitRendererPluginOptions.fonts;
        _asyncToGenerator(/*#__PURE__*/_regeneratorRuntime().mark(function _callee() {
          var background, clearColor, dpr, firstFrame, tmpVec3, tmpQuat, _drawFrame;
          return _regeneratorRuntime().wrap(function _callee$(_context) {
            while (1) switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return Promise.all(fonts.map(function (_ref2) {
                  var name = _ref2.name,
                    url = _ref2.url;
                  return _this.fontLoader.loadFont(CanvasKit, name, url);
                }));
              case 2:
                background = _this.context.config.background;
                clearColor = gLite.parseColor(background); // scale all drawing operations by the dpr
                // @see https://www.html5rocks.com/en/tutorials/canvas/hidpi/
                dpr = _this.context.contextService.getDPR();
                surface.getCanvas().scale(dpr, dpr);
                firstFrame = Date.now();
                tmpVec3 = glMatrix.vec3.create();
                tmpQuat = glMatrix.quat.create();
                _drawFrame = function drawFrame(canvas) {
                  if (_this.destroyed || surface.isDeleted()) {
                    return;
                  }
                  canvas.save();
                  _this.applyCamera(canvas, _this.context.camera, tmpVec3, tmpQuat);
                  canvas.clear(CanvasKit.Color4f(Number(clearColor.r) / 255, Number(clearColor.g) / 255, Number(clearColor.b) / 255, Number(clearColor.alpha)));
                  _this.drawAnimations(canvas, firstFrame);
                  _this.drawParticles(canvas);
                  _this.drawWithSurface(canvas, renderingContext.root);

                  // pop restore stack, eg. root -> parent -> child
                  _this.restoreStack.forEach(function () {
                    canvas.restore();
                  });
                  // clear restore stack
                  _this.restoreStack = [];
                  canvas.restore();

                  // capture here since we don't preserve drawing buffer
                  if (_this.enableCapture && _this.resolveCapturePromise) {
                    var _this$captureOptions = _this.captureOptions,
                      _this$captureOptions$ = _this$captureOptions.type,
                      type = _this$captureOptions$ === void 0 ? 'image/png' : _this$captureOptions$,
                      _this$captureOptions$2 = _this$captureOptions.encoderOptions,
                      encoderOptions = _this$captureOptions$2 === void 0 ? 1 : _this$captureOptions$2;
                    var typeMap = {
                      'image/png': CanvasKit.ImageFormat.PNG,
                      'image/jpeg': CanvasKit.ImageFormat.JPEG,
                      'image/webp': CanvasKit.ImageFormat.WEBP,
                      'image/bmp': undefined
                    };

                    // TODO: transparent image
                    var snapshot = surface.makeImageSnapshot();
                    var bytes = snapshot.encodeToBytes(typeMap[type], encoderOptions);
                    var blobObj = new Blob([bytes], {
                      type: type
                    });
                    _this.resolveCapturePromise(window.URL.createObjectURL(blobObj));
                    _this.enableCapture = false;
                    _this.captureOptions = undefined;
                    _this.resolveCapturePromise = undefined;
                  }
                  surface.requestAnimationFrame(_drawFrame);
                };
                surface.requestAnimationFrame(_drawFrame);
              case 11:
              case "end":
                return _context.stop();
            }
          }, _callee);
        }))();
      });
      renderingService.hooks.destroy.tap(CanvaskitRendererPlugin.tag, function () {
        var canvasKitContext = _this.context.contextService.getContext();
        var surface = canvasKitContext.surface;
        _this.animations.forEach(function (_ref3) {
          var animation = _ref3.animation;
          animation["delete"]();
        });
        _this.animations = [];
        _this.particlesList.forEach(function (_ref4) {
          var particles = _ref4.particles;
          particles["delete"]();
        });
        _this.particlesList = [];
        _this.fontLoader.destroy();
        surface.deleteLater();
        _this.destroyed = true;
      });
    }
  }, {
    key: "applyCamera",
    value: function applyCamera(canvas, camera, tmpVec3, tmpQuat) {
      var transform = camera.getOrthoMatrix();
      var _mat4$getTranslation = glMatrix.mat4.getTranslation(tmpVec3, transform),
        _mat4$getTranslation2 = _slicedToArray(_mat4$getTranslation, 2),
        tx = _mat4$getTranslation2[0],
        ty = _mat4$getTranslation2[1];
      var _mat4$getScaling = glMatrix.mat4.getScaling(tmpVec3, transform),
        _mat4$getScaling2 = _slicedToArray(_mat4$getScaling, 2),
        sx = _mat4$getScaling2[0],
        sy = _mat4$getScaling2[1];
      var rotation = glMatrix.mat4.getRotation(tmpQuat, transform);
      var _getEuler = gLite.getEuler(tmpVec3, rotation),
        _getEuler2 = _slicedToArray(_getEuler, 3),
        euz = _getEuler2[2];
      var rot = gLite.rad2deg(euz);
      canvas.translate(tx, ty);
      canvas.rotate(rot, 0, 0);
      canvas.scale(sx, sy);
    }
  }, {
    key: "drawAnimations",
    value: function drawAnimations(canvas, firstFrame) {
      var _this2 = this;
      var rectLeft = 0;
      var rectTop = 1;
      var rectRight = 2;
      var rectBottom = 3;
      this.animations.forEach(function (_ref5, i) {
        var duration = _ref5.duration,
          animation = _ref5.animation,
          bounds = _ref5.bounds;
        if (animation.isDeleted()) {
          _this2.animations.splice(i, 1);
        }
        var seek = (Date.now() - firstFrame) / duration % 1.0;
        var damage = animation.seek(seek);
        if (damage[rectRight] > damage[rectLeft] && damage[rectBottom] > damage[rectTop]) {
          animation.render(canvas, bounds);
        }
      });
    }
  }, {
    key: "drawParticles",
    value: function drawParticles(canvas) {
      var _this3 = this;
      this.particlesList.forEach(function (_ref6, i) {
        var particles = _ref6.particles,
          onFrame = _ref6.onFrame;
        if (particles.isDeleted()) {
          _this3.particlesList.splice(i, 1);
        }
        canvas.save();
        if (onFrame) {
          onFrame(canvas);
        }
        particles.update(Date.now() / 1000.0);
        particles.draw(canvas);
        canvas.restore();
      });
    }
  }, {
    key: "drawWithSurface",
    value: function drawWithSurface(canvas, object) {
      var _object$sortable,
        _this4 = this;
      if (object.isVisible() && !object.isCulled()) {
        this.renderDisplayObject(object, canvas);
      }
      var sorted = ((_object$sortable = object.sortable) === null || _object$sortable === void 0 || (_object$sortable = _object$sortable.sorted) === null || _object$sortable === void 0 ? void 0 : _object$sortable.length) > 0 ? object.sortable.sorted : object.childNodes;

      // should account for z-index
      sorted.forEach(function (child) {
        _this4.drawWithSurface(canvas, child);
      });
    }
  }, {
    key: "generatePattern",
    value: function generatePattern(object, pattern) {
      var _this5 = this;
      var _getContext = this.context.contextService.getContext(),
        surface = _getContext.surface,
        CanvasKit = _getContext.CanvasKit;
      var image = pattern.image,
        repetition = pattern.repetition,
        transform = pattern.transform;
      var src;
      if (util.isString(image)) {
        var imageCache = this.context.imagePool.getImageSync(image, object, function () {
          // set dirty rectangle flag
          object.renderable.dirty = true;
          _this5.context.renderingService.dirtify();
        });
        src = imageCache === null || imageCache === void 0 ? void 0 : imageCache.img;
      } else if (image.nodeName === 'rect') ; else {
        // @ts-ignore
        src = image;
      }
      if (src) {
        // TODO: check image.complete
        // WebGL: INVALID_VALUE: texImage2D: invalid image
        var decoded = surface.makeImageFromTextureSource(src);
        if (decoded) {
          var tx;
          var ty;
          if (repetition === 'repeat') {
            tx = CanvasKit.TileMode.Repeat;
            ty = CanvasKit.TileMode.Repeat;
          } else if (repetition === 'repeat-x') {
            tx = CanvasKit.TileMode.Repeat;
            ty = CanvasKit.TileMode.Decal;
          } else if (repetition === 'repeat-y') {
            tx = CanvasKit.TileMode.Decal;
            ty = CanvasKit.TileMode.Repeat;
          } else if (repetition === 'no-repeat') {
            tx = CanvasKit.TileMode.Decal;
            ty = CanvasKit.TileMode.Decal;
          }
          var mat;
          // @see https://developer.mozilla.org/en-US/docs/Web/API/CanvasPattern/setTransform
          if (transform) {
            mat = gLite.parsedTransformToMat4(gLite.parseTransform(transform), new gLite.DisplayObject({}));
          } else {
            mat = glMatrix.mat4.identity(glMatrix.mat4.create());
          }
          var _object$getGeometryBo = object.getGeometryBounds(),
            min = _object$getGeometryBo.min;
          var _pattern = decoded.makeShaderCubic(tx, ty, 1 / 3, 1 / 3, [mat[0], mat[4], mat[12] + min[0], mat[1], mat[5], mat[13] + min[1], 0, 0, 1]);
          return _pattern;
        }
      }
    }
  }, {
    key: "generateGradient",
    value: function generateGradient(object, stroke) {
      var _getContext2 = this.context.contextService.getContext(),
        CanvasKit = _getContext2.CanvasKit;
      var bounds = object.getGeometryBounds();
      var width = bounds && bounds.halfExtents[0] * 2 || 0;
      var height = bounds && bounds.halfExtents[1] * 2 || 0;
      var min = bounds && bounds.min || [0, 0];
      if (stroke.type === gLite.GradientType.LinearGradient) {
        // @see https://fiddle.skia.org/c/@GradientShader_MakeLinear
        var _ref7 = stroke.value,
          angle = _ref7.angle,
          steps = _ref7.steps;
        var pos = [];
        var colors = [];
        steps.forEach(function (_ref8) {
          var offset = _ref8.offset,
            color = _ref8.color;
          if (offset.unit === gLite.UnitType.kPercentage) {
            pos.push(offset.value / 100);
          }
          var c = gLite.parseColor(color);
          colors.push(Number(c.alpha) === 0 ? 1 : Number(c.r) / 255, Number(c.alpha) === 0 ? 1 : Number(c.g) / 255, Number(c.alpha) === 0 ? 1 : Number(c.b) / 255, Number(c.alpha));
        });
        var _computeLinearGradien = gLite.computeLinearGradient([min[0], min[1]], width, height, angle),
          x1 = _computeLinearGradien.x1,
          y1 = _computeLinearGradien.y1,
          x2 = _computeLinearGradien.x2,
          y2 = _computeLinearGradien.y2;
        var gradient = CanvasKit.Shader.MakeLinearGradient([x1, y1], [x2, y2], new Float32Array(colors), pos, CanvasKit.TileMode.Mirror);
        return gradient;
      }
      if (stroke.type === gLite.GradientType.RadialGradient) {
        var _ref9 = stroke.value,
          cx = _ref9.cx,
          cy = _ref9.cy,
          _steps = _ref9.steps,
          size = _ref9.size;
        var _computeRadialGradien = gLite.computeRadialGradient([min[0], min[1]], width, height, cx, cy, size),
          x = _computeRadialGradien.x,
          y = _computeRadialGradien.y,
          r = _computeRadialGradien.r;
        var _pos = [];
        var _colors = [];
        _steps.forEach(function (_ref0) {
          var offset = _ref0.offset,
            color = _ref0.color;
          if (offset.unit === gLite.UnitType.kPercentage) {
            _pos.push(offset.value / 100);
          }
          var c = gLite.parseColor(color);
          _colors.push(new Float32Array([Number(c.alpha) === 0 ? 1 : Number(c.r) / 255, Number(c.alpha) === 0 ? 1 : Number(c.g) / 255, Number(c.alpha) === 0 ? 1 : Number(c.b) / 255, Number(c.alpha)]));
        });
        // @see https://fiddle.skia.org/c/@radial_gradient_test
        var _gradient = CanvasKit.Shader.MakeRadialGradient([x, y], r, _colors, _pos, CanvasKit.TileMode.Clamp // mirror for repetition
        );
        return _gradient;
      }
    }
  }, {
    key: "generateGradientsShader",
    value: function generateGradientsShader(object, fill) {
      var _this6 = this;
      var _getContext3 = this.context.contextService.getContext(),
        CanvasKit = _getContext3.CanvasKit;
      var gradientShaders = fill.map(function (gradient) {
        return _this6.generateGradient(object, gradient);
      });
      var previousShader = gradientShaders[0];
      for (var i = 1; i < gradientShaders.length; i++) {
        previousShader = CanvasKit.Shader.MakeBlend(
        // @see https://developer.mozilla.org/zh-CN/docs/Web/API/CanvasRenderingContext2D/globalCompositeOperation
        CanvasKit.BlendMode.SrcOver, previousShader, gradientShaders[i]);
      }
      return previousShader;
    }
  }, {
    key: "renderDisplayObject",
    value: function renderDisplayObject(object, canvas) {
      var _getContext4 = this.context.contextService.getContext(),
        CanvasKit = _getContext4.CanvasKit;

      // restore to its ancestor
      var parent = this.restoreStack[this.restoreStack.length - 1];
      if (parent && !(object.compareDocumentPosition(parent) & gLite.Node.DOCUMENT_POSITION_CONTAINS)) {
        canvas.restore();
        this.restoreStack.pop();
      }
      var renderer = this.rendererContributionFactory[object.nodeName];
      var _object$parsedStyle = object.parsedStyle,
        fill = _object$parsedStyle.fill,
        stroke = _object$parsedStyle.stroke,
        lineWidth = _object$parsedStyle.lineWidth,
        _object$parsedStyle$l = _object$parsedStyle.lineCap,
        lineCap = _object$parsedStyle$l === void 0 ? 'butt' : _object$parsedStyle$l,
        _object$parsedStyle$l2 = _object$parsedStyle.lineJoin,
        lineJoin = _object$parsedStyle$l2 === void 0 ? 'miter' : _object$parsedStyle$l2,
        lineDash = _object$parsedStyle.lineDash,
        lineDashOffset = _object$parsedStyle.lineDashOffset,
        miterLimit = _object$parsedStyle.miterLimit,
        opacity = _object$parsedStyle.opacity,
        fillOpacity = _object$parsedStyle.fillOpacity,
        strokeOpacity = _object$parsedStyle.strokeOpacity,
        shadowBlur = _object$parsedStyle.shadowBlur,
        shadowColor = _object$parsedStyle.shadowColor,
        clipPath = _object$parsedStyle.clipPath;

      // apply clipPath
      if (clipPath) {
        canvas.save();
        // save clip
        this.restoreStack.push(object);

        /**
         * Since there's no resetMatrix in CanvasKit, so clipPath cannot be saved alone.
         * @see https://api.skia.org/classSkCanvas.html#aba129108fc68dca01850faf73d5db148
         * @see https://fiddle.skia.org/c/@Canvas_clipPath
         */
        var d = gLite.convertToPath(clipPath, clipPath.getWorldTransform());
        var path = new gLite.Path({
          style: {
            d: d
          }
        });
        var skPath = generateSkPath(CanvasKit, path);
        var _clipPath$getPosition = clipPath.getPosition(),
          _clipPath$getPosition2 = _slicedToArray(_clipPath$getPosition, 2),
          tx = _clipPath$getPosition2[0],
          ty = _clipPath$getPosition2[1];
        // FIXME: account for local skew
        // const [ax, ay] = clipPath.getLocalSkew();
        var _clipPath$getScale = clipPath.getScale(),
          _clipPath$getScale2 = _slicedToArray(_clipPath$getScale, 2),
          sx = _clipPath$getScale2[0],
          sy = _clipPath$getScale2[1];
        var rot = clipPath.getEulerAngles();
        var m = gLite.fromRotationTranslationScale(rot, tx, ty, sx, sy);
        skPath.transform([m[0], m[3], m[6], m[1], m[4], m[7], m[2], m[5], m[8]]);
        // CanvasKit only support clip Path now.
        canvas.clipPath(skPath, CanvasKit.ClipOp.Intersect, true);
      }
      if (renderer) {
        var _fillPaint, _strokePaint, _shadowFillPaint, _shadowStrokePaint;
        canvas.save();

        /**
         * world transform
         */
        var _object$getPosition = object.getPosition(),
          _object$getPosition2 = _slicedToArray(_object$getPosition, 2),
          _tx = _object$getPosition2[0],
          _ty = _object$getPosition2[1];
        var _object$getScale = object.getScale(),
          _object$getScale2 = _slicedToArray(_object$getScale, 2),
          _sx = _object$getScale2[0],
          _sy = _object$getScale2[1];
        var _rot = object.getEulerAngles();
        var _object$getLocalSkew = object.getLocalSkew(),
          _object$getLocalSkew2 = _slicedToArray(_object$getLocalSkew, 2),
          ax = _object$getLocalSkew2[0],
          ay = _object$getLocalSkew2[1];

        // @see https://fiddle.skia.org/c/68b54cb7d3435f46e532e6d565a59c49
        canvas.translate(_tx, _ty);
        // FIXME
        canvas.skew(ax, ay);
        canvas.rotate(_rot, 0, 0);
        canvas.scale(_sx, _sy);
        var hasFill = fill && !fill.isNone;
        var hasStroke = stroke && !stroke.isNone;
        var hasShadow = shadowColor && shadowBlur > 0;
        var fillPaint = null;
        var strokePaint = null;
        var shadowFillPaint = null;
        var shadowStrokePaint = null;
        if (hasFill) {
          fillPaint = new CanvasKit.Paint();
          fillPaint.setAntiAlias(true);
          fillPaint.setStyle(CanvasKit.PaintStyle.Fill);
          // should not affect transparent
          if (gLite.isCSSRGB(fill) && !fill.isNone) {
            fillPaint.setColorComponents(Number(fill.r) / 255, Number(fill.g) / 255, Number(fill.b) / 255, Number(fill.alpha) * fillOpacity * opacity);
          } else if (Array.isArray(fill)) {
            fillPaint.setAlphaf(fillOpacity * opacity);
            var shader = this.generateGradientsShader(object, fill);
            fillPaint.setShader(shader);
            shader["delete"]();
          } else if (gLite.isPattern(fill)) {
            fillPaint.setAlphaf(fillOpacity * opacity);
            var _shader = this.generatePattern(object, fill);
            if (_shader) {
              fillPaint.setShader(_shader);
              _shader["delete"]();
            }
          }
        }
        if (hasStroke) {
          strokePaint = new CanvasKit.Paint();
          /**
           * stroke
           */
          if (lineWidth > 0) {
            strokePaint.setAntiAlias(true);
            strokePaint.setStyle(CanvasKit.PaintStyle.Stroke);
            if (gLite.isCSSRGB(stroke) && !stroke.isNone) {
              strokePaint.setColor(CanvasKit.Color4f(Number(stroke.r) / 255, Number(stroke.g) / 255, Number(stroke.b) / 255, Number(stroke.alpha) * strokeOpacity * opacity));
            } else if (Array.isArray(stroke)) {
              strokePaint.setAlphaf(strokeOpacity * opacity);
              var _shader2 = this.generateGradientsShader(object, stroke);
              strokePaint.setShader(_shader2);
              _shader2["delete"]();
            } else if (gLite.isPattern(stroke)) {
              strokePaint.setAlphaf(strokeOpacity * opacity);
              var _shader3 = this.generatePattern(object, stroke);
              if (_shader3) {
                strokePaint.setShader(_shader3);
                _shader3["delete"]();
              }
            }
            strokePaint.setStrokeWidth(lineWidth);
            var STROKE_CAP_MAP = {
              butt: CanvasKit.StrokeCap.Butt,
              round: CanvasKit.StrokeCap.Round,
              square: CanvasKit.StrokeCap.Square
            };
            strokePaint.setStrokeCap(STROKE_CAP_MAP[lineCap]);
            var STROKE_JOIN_MAP = {
              bevel: CanvasKit.StrokeJoin.Bevel,
              round: CanvasKit.StrokeJoin.Round,
              miter: CanvasKit.StrokeJoin.Miter
            };
            strokePaint.setStrokeJoin(STROKE_JOIN_MAP[lineJoin]);
            if (!util.isNil(miterLimit)) {
              strokePaint.setStrokeMiter(miterLimit);
            }
            if (lineDash) {
              strokePaint.setPathEffect(CanvasKit.PathEffect.MakeDash(lineDash, lineDashOffset || 0));
            }
          }
        }
        if (hasFill && hasShadow) {
          shadowFillPaint = fillPaint.copy();
          shadowFillPaint.setAntiAlias(true);
          shadowFillPaint.setColor(CanvasKit.Color4f(Number(shadowColor.r) / 255, Number(shadowColor.g) / 255, Number(shadowColor.b) / 255, Number(shadowColor.alpha)));
          var blurSigma = (shadowBlur && shadowBlur || 0) / 2;
          shadowFillPaint.setMaskFilter(CanvasKit.MaskFilter.MakeBlur(CanvasKit.BlurStyle.Normal, blurSigma, false));
        }
        if (hasStroke && hasShadow) {
          shadowStrokePaint = strokePaint.copy();
          shadowStrokePaint.setAntiAlias(true);
          shadowStrokePaint.setColor(CanvasKit.Color4f(Number(shadowColor.r) / 255, Number(shadowColor.g) / 255, Number(shadowColor.b) / 255, Number(shadowColor.alpha)));
          var _blurSigma = (shadowBlur && shadowBlur || 0) / 2;
          shadowStrokePaint.setMaskFilter(CanvasKit.MaskFilter.MakeBlur(CanvasKit.BlurStyle.Normal, _blurSigma, false));
        }
        renderer.render(object, {
          fillPaint: fillPaint,
          strokePaint: strokePaint,
          shadowFillPaint: shadowFillPaint,
          shadowStrokePaint: shadowStrokePaint,
          canvas: canvas
        });
        (_fillPaint = fillPaint) === null || _fillPaint === void 0 || _fillPaint["delete"]();
        (_strokePaint = strokePaint) === null || _strokePaint === void 0 || _strokePaint["delete"]();
        (_shadowFillPaint = shadowFillPaint) === null || _shadowFillPaint === void 0 || _shadowFillPaint["delete"]();
        (_shadowStrokePaint = shadowStrokePaint) === null || _shadowStrokePaint === void 0 || _shadowStrokePaint["delete"]();
        canvas.restore();
      }

      // finish rendering, clear dirty flag
      object.renderable.dirty = false;
    }
  }, {
    key: "toDataURL",
    value: function () {
      var _toDataURL = _asyncToGenerator(/*#__PURE__*/_regeneratorRuntime().mark(function _callee2(options) {
        var _this7 = this;
        return _regeneratorRuntime().wrap(function _callee2$(_context2) {
          while (1) switch (_context2.prev = _context2.next) {
            case 0:
              // trigger re-render
              this.enableCapture = true;
              this.captureOptions = options;
              this.capturePromise = new Promise(function (resolve) {
                _this7.resolveCapturePromise = function (dataURL) {
                  resolve(dataURL);
                };
              });
              return _context2.abrupt("return", this.capturePromise);
            case 4:
            case "end":
              return _context2.stop();
          }
        }, _callee2, this);
      }));
      function toDataURL(_x) {
        return _toDataURL.apply(this, arguments);
      }
      return toDataURL;
    }()
  }]);
}();
CanvaskitRendererPlugin.tag = 'CanvaskitRenderer';

// const cdn = 'https://storage.googleapis.com/skia-cdn/misc/';

var FontLoader = /*#__PURE__*/function () {
  function FontLoader() {
    _classCallCheck(this, FontLoader);
    this.cache = {};
    this.fontMgrCache = {};
  }
  return _createClass(FontLoader, [{
    key: "loadFont",
    value: function () {
      var _loadFont = _asyncToGenerator(/*#__PURE__*/_regeneratorRuntime().mark(function _callee(CanvasKit, font, url) {
        var response, fontBuffer, fontSrc, typeface;
        return _regeneratorRuntime().wrap(function _callee$(_context) {
          while (1) switch (_context.prev = _context.next) {
            case 0:
              _context.next = 2;
              return fetch(url);
            case 2:
              response = _context.sent;
              _context.next = 5;
              return response.arrayBuffer();
            case 5:
              fontBuffer = _context.sent;
              fontSrc = CanvasKit.TypefaceFontProvider.Make();
              typeface = CanvasKit.Typeface.MakeFreeTypeFaceFromData(fontBuffer);
              fontSrc.registerFont(fontBuffer, font);
              this.cache[font] = {
                provider: fontSrc,
                typeface: typeface,
                fontBuffer: fontBuffer
              };
            case 10:
            case "end":
              return _context.stop();
          }
        }, _callee, this);
      }));
      function loadFont(_x, _x2, _x3) {
        return _loadFont.apply(this, arguments);
      }
      return loadFont;
    }()
  }, {
    key: "getTypefaceFontProvider",
    value: function getTypefaceFontProvider(font) {
      var _this$cache$font;
      return (_this$cache$font = this.cache[font]) === null || _this$cache$font === void 0 ? void 0 : _this$cache$font.provider;
    }
  }, {
    key: "getTypeface",
    value: function getTypeface(font) {
      var _this$cache$font2;
      return (_this$cache$font2 = this.cache[font]) === null || _this$cache$font2 === void 0 ? void 0 : _this$cache$font2.typeface;
    }
  }, {
    key: "getFontBuffer",
    value: function getFontBuffer(font) {
      var _this$cache$font3;
      return (_this$cache$font3 = this.cache[font]) === null || _this$cache$font3 === void 0 ? void 0 : _this$cache$font3.fontBuffer;
    }
  }, {
    key: "getOrCreateFontMgr",
    value: function getOrCreateFontMgr(CanvasKit, fontFamilies) {
      var _this = this;
      var cacheKey = fontFamilies.join(',');
      if (!this.fontMgrCache[cacheKey]) {
        var _CanvasKit$FontMgr;
        this.fontMgrCache[cacheKey] = (_CanvasKit$FontMgr = CanvasKit.FontMgr).FromData.apply(_CanvasKit$FontMgr, _toConsumableArray(fontFamilies.map(function (font) {
          return _this.getFontBuffer(font);
        })));
      }
      return this.fontMgrCache[cacheKey];
    }
  }, {
    key: "destroy",
    value: function destroy() {
      var _this2 = this;
      Object.keys(this.fontMgrCache).forEach(function (font) {
        _this2.fontMgrCache[font]["delete"]();
      });
      Object.keys(this.cache).forEach(function (key) {
        _this2.cache[key].provider["delete"]();
        _this2.cache[key].typeface["delete"]();
      });
      this.cache = {};
      this.fontMgrCache = {};
    }
  }]);
}();

var Plugin = /*#__PURE__*/function (_AbstractRendererPlug) {
  function Plugin() {
    var _this;
    var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    _classCallCheck(this, Plugin);
    _this = _callSuper(this, Plugin);
    _this.name = 'canvaskit-renderer';
    _this.options = options;
    return _this;
  }
  _inherits(Plugin, _AbstractRendererPlug);
  return _createClass(Plugin, [{
    key: "init",
    value: function init() {
      var _rendererContribution;
      var canvaskitRendererPluginOptions = _objectSpread({
        fonts: []
      }, this.options);
      var fontLoader = new FontLoader();
      var rendererContributionFactory = (_rendererContribution = {}, _defineProperty(_defineProperty(_defineProperty(_defineProperty(_defineProperty(_defineProperty(_defineProperty(_defineProperty(_defineProperty(_defineProperty(_rendererContribution, gLite.Shape.CIRCLE, new CircleRenderer()), gLite.Shape.ELLIPSE, new EllipseRenderer()), gLite.Shape.RECT, new RectRenderer()), gLite.Shape.IMAGE, new ImageRenderer(this.context)), gLite.Shape.TEXT, new TextRenderer(this.context, fontLoader)), gLite.Shape.LINE, new LineRenderer()), gLite.Shape.POLYLINE, new PolylineRenderer(this.context)), gLite.Shape.POLYGON, new PolygonRenderer(this.context)), gLite.Shape.PATH, new PathRenderer(this.context)), gLite.Shape.GROUP, undefined), _defineProperty(_defineProperty(_rendererContribution, gLite.Shape.HTML, undefined), gLite.Shape.MESH, undefined));
      this.addRenderingPlugin(new CanvaskitRendererPlugin(canvaskitRendererPluginOptions, rendererContributionFactory, fontLoader));
    }
  }, {
    key: "destroy",
    value: function destroy() {
      this.removeAllRenderingPlugins();
    }
  }, {
    key: "playAnimation",
    value: function playAnimation(name, jsonStr, bounds, assets) {
      return this.plugins[0].playAnimation(name, jsonStr, bounds, assets);
    }
  }, {
    key: "createParticles",
    value: function createParticles(jsonStr, onFrame, assets) {
      return this.plugins[0].createParticles(jsonStr, onFrame, assets);
    }
  }, {
    key: "toDataURL",
    value: function toDataURL(options) {
      return this.plugins[0].toDataURL(options);
    }
  }]);
}(gLite.AbstractRendererPlugin);

exports.Plugin = Plugin;
//# sourceMappingURL=index.js.map
